from enum import Enum


class UseType(str, Enum):
    """The use type that the vehicle is used for"""
    special_use = "Særlig anvendelse"
    private_driving = "Privat personkørsel"
    residence = "Beboelse"
    freight_transport_half_oms = "Godstransport (½ OMS.)"
    freight_transport_business = "Godstransport erhverv"
    freight_transport_private_and_business = "Godstransport privat/erhverv"
    freight_transport_private = "Godstransport privat"
    limousine_driving = "Limousinekørsel"
    public_authority_driving = "Kørsel for offentlig myndighed"
    freight_transport = "Godstransport"
    taxi_driving = "Taxikørsel"
    medical_transport = "Sygetransport"
    private_bus_driving = "Privat buskørsel"



from enum import Enum


class BodyType(str, Enum):
    """Vehicle body types"""
    cabriolet = "Cabriolet"
    coupe = "Coupe"
    hatchback = "Hatchback"
    mpw = "MPV"
    sedan = "Sedan"
    stationcar = "Stationcar"
    pickup = "Stationcar-pickup"



from __future__ import annotations
from pydantic import BaseModel
from typing import Optional
import datetime


class Insurance(BaseModel):
    """
    The Insurance object holds the insurance information about the vehicle.

    Possible attributes:

        Insurance.company (str): The insurance provider.

        Insurance.is_active (bool): Whether the insurance is active.

        Insurance.number (str): The insurance number, typically returns None.

        Insurance.created (datetime): The date when the insurance was created.
    """
    company: str
    is_active: bool
    number: Optional[int]
    created: datetime.datetime



from .models.insurance import Insurance
from .models.vehicle import Vehicle
from .models.propulsion_type import PropulsionType
from .models.vehicle_type import VehicleType
from .models.body_type import BodyType
from .models.use_type import UseType
from .dmr import DMR
from .converter import Converter



from .body_type import BodyType
from .insurance import Insurance
from .propulsion_type import PropulsionType
from .use_type import UseType
from .vehicle_type import VehicleType
from .vehicle import Vehicle



# xpaths for the different elements on the site
XPATHS = {
    "other": {
        "token": "/html/body/div[2]/div/div[1]/div[2]/form/input",
        "token_url": "/html/body/div[2]/div/div[1]/div[2]/form",
    },
    
    "page_1": {
        "make_model_variant": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[1]/div/div[2]/span[2]",
        "vin": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[1]/div/div[1]/span[2]",
        "type": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[1]/div/div[3]/span[2]",
        "last_update": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[1]/div/div[4]/span[2]",
        "registration_number": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[2]/div/div[1]/span[2]",
        "first_registration": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[4]/div/div[2]/div[1]/div[2]/span",
        "use": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[2]/div/div[3]/span[2]",
        "vehicle_id": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[4]/div/div[1]/div[1]/div[2]/span",
        "color": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[4]/div/div[1]/div[3]/div[2]/span",
        "model_year": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[4]/div/div[1]/div[4]/div[2]/span",
    },
    
    "page_2": {
        "total_weight": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[2]/div[2]/span[1]",
        "vehicle_weight": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[1]/div[3]/div[2]/span[1]",
        "propulsion": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[3]/div/div[1]/div[2]/div[2]/span",
        "tow_bar": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[2]/div/div[2]/div[3]/div[2]/span",
        "fuel_consumption": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[5]/div/div[1]/div[1]/div[2]/span[1]",
        "cylinders": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[3]/div/div[2]/div[3]/div[2]/span",
        "plugin_hybrid": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[3]/div/div[1]/div[3]/div[2]/span",
        "electricity_consumption": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[5]/div/div[1]/div[2]/div[2]/span",
        "electric_range": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[5]/div/div[2]/div[1]/div[2]/span",
        "battery_capacity": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[5]/div/div[2]/div[2]/div[2]/span",
        "body_type": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[6]/div/div[1]/div[1]/div[2]/span",
        "particle_filter": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[8]/div/div[1]/fieldset/div[6]/div[2]/span",
        "doors": "/html/body/div[2]/div/div[1]/div[2]/div[3]/div/div[2]/div[1]/div[6]/div/div[1]/div[3]/div[2]/span",
    },
    
    "page_3": None,
    
    "page_4": {
        "company": '//*[@id="lblSelskab"]',
        "is_active": '//*[@id="lblStatus"]',
        "number": '//*[@id="lblBevisNummer"]',
        "created": '//*[@id="lblbevisdato"]',
    },
}


from random import choice

USER_AGENTS = [
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.0; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0",
    "Mozilla/5.0 (Linux; Android 5.0.2; LG-V410/V41020b Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/34.0.1847.118 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2503.0 Safari/537.36",
    "Mozilla/5.0 (X11; Fedora; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 8_1_1 like Mac OS X) AppleWebKit/600.1.4 (KHTML, like Gecko) Version/8.0 Mobile/12B435 Safari/600.1.4",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.130 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:28.0) Gecko/20100101 Firefox/28.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:36.0) Gecko/20100101 Firefox/36.0",
    "Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.157 Safari/537.36",
    "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; InfoPath.3; .NET4.0C; .NET4.0E)",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/45.0.2454.85 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/45.0.2454.85 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.115 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.2; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; MDDRJS; rv:11.0) like Gecko",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Maxthon/4.4.6.2000 Chrome/30.0.1599.101 Safari/537.36",
    "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.3; WOW64; Trident/6.0)",
    "Mozilla/5.0 (Linux; Android 5.1.1; SAMSUNG SM-G920T Build/LMY47X) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/3.2 Chrome/38.0.2125.102 Mobile Safari/537.36",
    "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; .NET CLR 3.5.30729; .NET CLR 3.0.30729; Media Center PC 6.0; .NET4.0C; .NET4.0E; InfoPath.3; MS-RTC LM 8)",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2503.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.91 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 4.4.3; KFTHWI Build/KTU84M) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/34.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/45.0.2454.85 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 4.4.3; KFSAWI Build/KTU84M) AppleWebKit/537.36 (KHTML, like Gecko) Silk/44.1.81 like Chrome/44.0.2403.128 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1; rv:32.0) Gecko/20100101 Firefox/32.0",
    "Mozilla/5.0 (Linux; Android 4.4.2; SM-T230NU Build/KOT49H) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.133 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 4.2.2; SM-T110 Build/JDQ39) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/45.0.2454.84 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 5.0.1; SAMSUNG SM-N910T Build/LRX22C) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/2.1 Chrome/34.0.1847.76 Mobile Safari/537.36",
    "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; .NET4.0C; .NET4.0E)",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36",
    "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Win64; x64; Trident/7.0)",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.10; rv:33.0) Gecko/20100101 Firefox/33.0",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.154 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.99 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; WOW64; rv:38.0) Gecko/20100101 Firefox/38.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36",
    "Mozilla/5.0 (X11; CrOS armv7l 6946.86.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.94 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:38.0) Gecko/20100101 Firefox/38.0 SeaMonkey/2.35",
    "Mozilla/5.0 (Linux; Android 4.4.2; SM-T330NU Build/KOT49H) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/45.0.2454.84 Safari/537.36",
    "Mozilla/5.0 (iPad; CPU OS 6_0_1 like Mac OS X) AppleWebKit/536.26 (KHTML, like Gecko) Version/6.0 Mobile/10A8426 Safari/8536.25",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.157 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 5.0.2; LG-V410 Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/45.0.2454.84 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.63 Safari/537.36 TheWorld 6",
    "Mozilla/5.0 (iPad; CPU OS 8_1 like Mac OS X) AppleWebKit/600.1.4 (KHTML, like Gecko) GSA/7.0.55539 Mobile/12B410 Safari/600.1.4",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/600.2.5 (KHTML, like Gecko) Version/8.0 Safari/600.1.25",
    "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0; EIE10;ENUSWOL)",
]

def get_user_agent():
    return choice(USER_AGENTS)


from enum import Enum


class VehicleType(str, Enum):
    car = "Personbil"
    van = "Varebil"
    truck = "Lastbil"
    tractor = "Traktor"
    motorcycle = "Motorcykel"
    moped = "Lille knallert"
    large_moped = "Stor knallert"
    trailer = "Påhængsvogn"
    bus = "Stor personbil"  # Busses are registered as "Stor personbil"



from dmr import Converter


def test_km_to_miles():
    assert Converter.km_to_miles(17) == 10.56


def test_miles_to_km():
    assert Converter.miles_to_km(17) == 27.36


def test_kg_to_lb():
    assert Converter.kg_to_lb(17) == 37.48


def test_lb_to_kg():
    assert Converter.lb_to_kg(17) == 7.71



class InvalidLicensePlate(Exception):
    """Exception raised for invalid license plate.

    Attributes:
        message -- explanation of the error
    """
    def __init__(self, message="Invalid license plate was given"):
        self.message = message
        super().__init__(self.message)

class UnexpectedSiteContent(Exception):
    """Exception raised when the dmr site has been changed and dmr.py doesn't support the change.

    Attributes:
        message -- explanation of the error
    """
    def __init__(self, message="DMR content has been updated, and the used version of dmr.py doesn't support the change."):
        self.message = message
        super().__init__(self.message)

class MissingToken(Exception):
    """Exception raised when the dmr site haven't given a token where it was expected to be.

    Attributes:
        message -- explanation of the error
    """
    def __init__(self, message="DMR token was not found, this may be a temporary error or could be a change in the way the DMR site handles requests that the current version of dmr.py doesn't support."):
        self.message = message
        super().__init__(self.message)


from dmr import DMR
import asyncio


def test_validate_license_plate():
    """Test license plate validator"""
    is_license_plate = DMR.validate_license_plate("GGGGGGG")
    assert is_license_plate == True

    is_license_plate = DMR.validate_license_plate("ThisIsNotReal")
    assert is_license_plate == False


def test_invalid_license_plate():
    dmr_obj = DMR.get_by_plate(license_plate="GGGGGGG")
    assert dmr_obj is None

def test_invalid_license_plate_async():
    loop = asyncio.new_event_loop()
    dmr_obj = loop.run_until_complete(DMR.get_by_plate_async("GGGGGGG"))
    assert dmr_obj is None



from aiohttp import ClientSession
from lxml.html import fromstring
from typing import Optional, Tuple
from .headers import get_headers
from .extract_data import page_1, page_2, page_4, get_token_and_url


async def get_token(session) -> Tuple[str, str]:
    """Get dmrFormToken"""
    async with session.get("https://motorregister.skat.dk/dmr-kerne/koeretoejdetaljer/visKoeretoej", headers=get_headers(), allow_redirects=True) as resp:
        content = await resp.text()
    source = fromstring(content)
    token, url = get_token_and_url(source)
    return token, url


async def scrape_async(license_plate: str) -> Optional[dict]:

    async with ClientSession() as session:
        # Get dmrFormToken required to make site requests and get url to post data
        token, new_url = await get_token(session)

        payload = {
            "dmrFormToken": token,
            "soegeord": license_plate,
            "soegekriterie:": "REGISTRERINGSNUMMER",
            new_url: "Søg"
        }

        async with session.post('https://motorregister.skat.dk' + new_url, data=payload, headers=get_headers({"Referer":"https://motorregister.skat.dk" + new_url}), allow_redirects=True) as resp:
            content = await resp.text()

        if "Ingen køretøjer fundet." in content:
            # license plate doesn't exist
            return None

        # Page 1 scrape
        source = fromstring(content)
        data = page_1(source)

        new_url_without_last_16_chars = new_url[:-16]

        # Page 2 scrape
        async with session.get(str(resp.url) + "&_eventId=customPage&_pageIndex=1", headers=get_headers({"Referer":"https://motorregister.skat.dk" + new_url_without_last_16_chars}), allow_redirects=True) as resp:
            content = await resp.text()
        source = fromstring(content)
        data.update(page_2(source))

        # Page 4 scrape
        async with session.get(str(resp.url) + "&_eventId=customPage&_pageIndex=3", headers=get_headers({"Referer":"https://motorregister.skat.dk" + new_url_without_last_16_chars}), allow_redirects=True) as resp:
            content = await resp.text()
        if "Ingen forsikring" not in content:
            source = fromstring(content)
            data.update(page_4(source))
        else:
            data.update({"insurance": None})

    return data



from .helper_methods import get_object_for_comparison, validate_object


def test_dmr_object_normal_car():
    dmr, expected = get_object_for_comparison("cw87553", False)
    validate_object(dmr, expected)


def test_dmr_object_electric_car():
    dmr, expected = get_object_for_comparison("ap43115", False)
    validate_object(dmr, expected)


def test_dmr_object_plugin_hybrid_car():
    dmr, expected = get_object_for_comparison("dd24506", False)
    validate_object(dmr, expected)


def test_dmr_object_hydrogen_car():
    dmr, expected = get_object_for_comparison("cy41511", True)
    validate_object(dmr, expected)


def test_dmr_object_van():
    dmr, expected = get_object_for_comparison("ap22698", True)
    validate_object(dmr, expected)


def test_dmr_object_motorcycle():
    dmr, expected = get_object_for_comparison("aw85002", True)
    validate_object(dmr, expected)



from datetime import datetime
from typing import Optional
from .propulsion_type import PropulsionType
from .vehicle_type import VehicleType
from .body_type import BodyType
from .use_type import UseType
from .insurance import Insurance
from pydantic import (
    BaseModel,
    create_model,
    ConfigDict,
)


class Vehicle(BaseModel):
    """This class represents a vehicle returned from DMR.

    Attributes:
        make:                      The make/manufacturer of the vehicle
        model:                     The vehicle model, if the vehicle were to be a Suzuki Swift, then the model would be "Swift\"
        variant:                   This is the variant of the model, a VW Golf GTI, would have the GTI variation of the Golf model
        type:                      The type of vehicle
        color:                     The color of the vehicle, typically None
        total_weight:              The maximum weight the vehicle can be when fully loaded, value is in KG
        vin:                       The vehicles identification number (VIN)
        last_update:               The last vehicle update, this could be the latest re-registration
        registration_number:       The vehicles license plate
        first_registration:        The date where the vehicle was first registered
        use:                       The type of usage the vehicle is being used for
        model_year:                The model year of the vehicle
        vehicle_weight:            The weight of the vehicle. The value returned is in KG
        propulsion:                The type fuel needed to use the vehicle
        tow_bar:                   Whether the vehicle has a tow bar
        fuel_consumption:          The fuel consumption of the vehicle, the value can be km/L if the vehicle runs on Benzin or Diesel
        cylinders:                 The amount of cylinders the vehicle has. A return value of None doesn't necessarily mean that the vehicle has no cylinders
        plugin_hybrid:             Whether the vehicle is a Plug-In hybrid
        electricity_consumption:   The electric consumption of the vehicle, the value can be Wh/km
        electric_range:            The electric range of the vehicle, value is in km
        battery_capacity:          The battery capacity of the vehicle, value is in kWh
        body_type:                 The body type of the vehicle
        vehicle_id:                The vehicle ID in the DMR system
        doors:                     The amount of doors the vehicle has
        particle_filter:           Whether the vehicle has a particle filter
        insurance:                 An insurance object
    """
    model_config = ConfigDict(protected_namespaces=())
    make: str
    model: str
    variant: str
    type: VehicleType
    color: Optional[str]
    total_weight: int
    vin: str
    last_update: datetime
    registration_number: str
    first_registration: datetime
    use: UseType
    model_year: Optional[int]
    vehicle_weight: Optional[int]
    propulsion: PropulsionType
    tow_bar: bool
    fuel_consumption: Optional[float]
    cylinders: Optional[int]
    plugin_hybrid: bool
    electricity_consumption: Optional[float]
    electric_range: Optional[float]
    battery_capacity: Optional[float]
    body_type: Optional[BodyType]
    vehicle_id: int
    doors: Optional[int]
    particle_filter: Optional[bool]
    insurance: Optional[create_model("Insurance", __base__=Insurance)]





ONE_MILE_IN_KM = 0.6213711922
ONE_KM_IN_MILES = 1.609344

ONE_POUND_IN_KG = 2.204623
ONE_KG_IN_POUNDS = 0.4535923


class Converter:

    @classmethod
    def km_to_miles(cls, km: int, decimals: int = 2):
        """Convert Kilometers to Miles"""
        miles = km * ONE_MILE_IN_KM
        return round(miles, decimals)

    @classmethod
    def miles_to_km(cls, miles: int, decimals: int = 2):
        """Convert Miles to Kilometers"""
        km = miles * ONE_KM_IN_MILES
        return round(km, decimals)

    @classmethod
    def kg_to_lb(cls, kg: int, decimals: int = 2):
        """Convert Kilograms to Pounds"""
        lb = kg * ONE_POUND_IN_KG
        return round(lb, decimals)

    @classmethod
    def lb_to_kg(cls, lb: int, decimals: int = 2):
        """Convert Pounds to Kilograms"""
        kg = lb * ONE_KG_IN_POUNDS
        return round(kg, decimals)



from lxml.html import HtmlElement
from .xpaths import XPATHS
from datetime import datetime
from .errors import MissingToken
from typing import Tuple


def get_value_from_xpath(source: HtmlElement, xpath_string: str):
    """Extracts the value from the XPath and turns their value into python understandable values"""
    content = source.xpath(xpath_string)[0].text_content()
    
    if content.startswith("-") or content == "Ukendt":
        return None
    
    if content in ["Nej", "Ja"]:
        return True if content == "Ja" else False
    
    if content.isnumeric():
        content = int(content)

    if isinstance(content, str):
        content = content.strip()

    return content


def get_token_and_url(source: HtmlElement) -> Tuple[str, str]:
    try:
        token = source.xpath(XPATHS["other"]["token"])[0].get("value")
        url = source.xpath(XPATHS["other"]["token_url"])[0].get("action")
        return token, url
    except (TypeError, KeyError):
        raise MissingToken("The scraper wasn't able to get a token from motorregister.skat.dk, the site may have changed.")


# Landing page when looking up a license plate, the page title is "1. Køretøj"
def page_1(source: HtmlElement) -> dict:
    data = {}
    make, model, variant = get_value_from_xpath(source, XPATHS["page_1"]["make_model_variant"]).split(",", 2)
    # Make make/model/variant names prettier
    data["make"] = make.strip().capitalize()
    data["model"] = model.strip().capitalize()
    data["variant"] = variant.strip().capitalize()
    data["vin"] = get_value_from_xpath(source, XPATHS["page_1"]["vin"])
    data["type"] = get_value_from_xpath(source, XPATHS["page_1"]["type"])
    data["last_update"] = datetime.strptime(get_value_from_xpath(source, XPATHS["page_1"]["last_update"]).split(" ")[-1], "%d-%m-%Y")
    data["registration_number"] = get_value_from_xpath(source, XPATHS["page_1"]["registration_number"])
    data["first_registration"] = datetime.strptime(get_value_from_xpath(source, XPATHS["page_1"]["first_registration"]).split(" ")[-1], "%d-%m-%Y")
    data["use"] = get_value_from_xpath(source, XPATHS["page_1"]["use"])
    data["vehicle_id"] = get_value_from_xpath(source, XPATHS["page_1"]["vehicle_id"])
    data["color"] = get_value_from_xpath(source, XPATHS["page_1"]["color"])
    data["model_year"] = get_value_from_xpath(source, XPATHS["page_1"]["model_year"])
    return data


# page title is "2. Tekniske oplysninger"
def page_2(source: HtmlElement) -> dict:
    data = {}
    data["total_weight"] = get_value_from_xpath(source, XPATHS["page_2"]["total_weight"])
    data["vehicle_weight"] = get_value_from_xpath(source, XPATHS["page_2"]["vehicle_weight"])
    data["propulsion"] = get_value_from_xpath(source, XPATHS["page_2"]["propulsion"])
    data["tow_bar"] = get_value_from_xpath(source, XPATHS["page_2"]["tow_bar"])
    fuel_consumption = get_value_from_xpath(source, XPATHS["page_2"]["fuel_consumption"])
    data["fuel_consumption"] = float(fuel_consumption.replace(",", ".")) if fuel_consumption is not None else None
    data["cylinders"] = get_value_from_xpath(source, XPATHS["page_2"]["cylinders"])
    data["plugin_hybrid"] = get_value_from_xpath(source, XPATHS["page_2"]["plugin_hybrid"])
    electricity_consumption = get_value_from_xpath(source, XPATHS["page_2"]["electricity_consumption"])
    data["electricity_consumption"] = float(electricity_consumption.replace(",", ".")) if electricity_consumption is not None else None
    electric_range = get_value_from_xpath(source, XPATHS["page_2"]["electric_range"])
    data["electric_range"] = float(electric_range.replace(",", ".")) if electric_range is not None else None
    battery_capacity = get_value_from_xpath(source, XPATHS["page_2"]["battery_capacity"])
    data["battery_capacity"] = float(battery_capacity.replace(",", ".")) if battery_capacity is not None else None
    data["body_type"] = get_value_from_xpath(source, XPATHS["page_2"]["body_type"])
    data["particle_filter"] = get_value_from_xpath(source, XPATHS["page_2"]["particle_filter"])
    data["doors"] = get_value_from_xpath(source, XPATHS["page_2"]["doors"])
    return data

# Page 3: "Syn"


# Page 4: "Forsikring"
def page_4(source: HtmlElement) -> dict:
    data = {"insurance": {}}
    data["insurance"]["company"] = get_value_from_xpath(source, XPATHS["page_4"]["company"])
    data["insurance"]["is_active"] = True if get_value_from_xpath(source, XPATHS["page_4"]["is_active"]) == "Aktiv" else False
    number = get_value_from_xpath(source, XPATHS["page_4"]["number"])
    data["insurance"]["number"] = number if number != "N/A" else None
    data["insurance"]["created"] = datetime.strptime(get_value_from_xpath(source, XPATHS["page_4"]["created"]), "%d-%m-%Y")
    return data



from .user_agent import get_user_agent


def get_headers(data: dict = None):
    headers = {
        'User-Agent': get_user_agent(),
    }

    if data:
        headers = headers.update(data)

    return headers



from __future__ import annotations
from typing import Optional
from dmr.models import Vehicle
from re import search
from dmr.utils import (
    scrape_async,
    scrape,
    errors,
)


class DMR:

    @classmethod
    def validate_license_plate(cls, license_plate: str) -> bool:
        """Checks if the given string can be a license plate, this check is also run prior to the webscraping.

        Args:
            license_plate (str): The license plate to check

        Returns:
            bool: Returns True if the given license plate could be a license plate, or False if not.
        """

        match = search(r"^[A-ZÆØÅ0-9]{2,7}$", license_plate.upper())
        if match:
            return True
        else:
            return False


    @classmethod
    def get_by_plate(cls, license_plate: str) -> Optional[Vehicle]:
        """Get data from DMR by license plate.

        Args:
            license_plate (str): The license plate to look up. Not required if license plate was passed into the DMR object.

        Raises:
            InvalidLicensePlate: Invalid license plate was given

        Returns:
            DMR: DMR object is returned, check https://github.com/j4asper/dmr.py/wiki for more information.
        """
        
        if not cls.validate_license_plate(license_plate):
            raise errors.InvalidLicensePlate("Invalid license plate. license plate length should be between 2 and 7 letters and/or digits.")

        data = scrape(license_plate=license_plate)
        if data:
            return Vehicle(**data)
        else:
            return None


    @classmethod
    async def get_by_plate_async(cls, license_plate: str) -> Optional[Vehicle]:
        """Get data from DMR asynchronously by license plate.

        Args:
            license_plate (str): The license plate to look up.

        Raises:
            InvalidLicensePlate: Invalid license plate was given

        Returns:
            DMR: DMR object is returned, check https://github.com/j4asper/dmr.py/wiki for more information.
        """

        if not cls.validate_license_plate(license_plate):
            raise errors.InvalidLicensePlate(
                "Invalid license plate. license plate length should be between 2 and 7 letters and/or digits.")

        data = await scrape_async(license_plate=license_plate)
        if data:
            return Vehicle(**data)
        else:
            return None



from enum import Enum


class PropulsionType(str, Enum):
    """Vehicle propulsion type, the kind of fuel the car uses"""
    diesel = "Diesel"
    benzin = "Benzin"
    electric = "El"
    hydrogen = "Brint"
    f_gas = "F-Gas"
    petroleum = "Petroleum"
    n_gas = "N-Gas"
    methanol = "Metanol"
    ethanol = "Ætanol"
